
// CBG ANSI terminal (VT 100).

#include <stdio.h>

const version = 5;

const leftmarg = 30;
const wcols = 770;   // Rhs margin wrap point
const topmarg  = 6;
const lineheight = 10;
const homex = leftmarg;
const homey = 16;
const lines = 25;
const cusymax = 250;  // lines times lineheight
const char Xon = 0x11;
const char Xoff = 0x13;
extern int cusx, cusy;
extern int cusx, cusy;
extern int scrolllock, holding;
extern char *addr;
extern char exitchar;
extern void kyread();
extern bell();
extern lockmark();
extern scroll();
extern crtcinit();
extern uartinit();
extern custime();
extern cusout();
extern crtcclr();
extern move();
extern draw();
extern paintstr();
extern kybdinit();
extern kybdrd();
extern sgetc();
extern sputc();
extern setbaud();
extern paintchar();

extern char notmoved, escpending, pad, graphpend, locked, cmdreq, mode;
extern int curmargin;


#define cursorup(X) if (cusy > 0) cusy = cusy - lineheight;
#define cursorleft(X) if (cusx >0) cusx = cusx - 8;
#define cursorright(X) cusx = cusx + 8;
#define cursortab(X) cusout(); cusx = (cusx & 0xFC0) + 64;

void printf(s, x) char *s;
{ 
  char c;
  while(*s)
    {
    c = *(s++);
    if (c == '%')
      {
      c = *(s++);
      if (c == 'x') paintx2(x);
      else sawrch('0' + x);
      }
    else sawrch(c);
    if (c == '\n') sawrch(13);
    }
}


int main()
{
  notmoved   = 0;
  escpending = 0;
  graphpend  = 0;
  locked     = 0;
  cmdreq     = 0;
  holding    = 0;
  crtcinit();
  clrhome();
  screen_toggle();
  clrhome();
  kybdinit();
  setbaud('M'); /* default to 9600 baud */
  paintstr("CBG Ansi Term 92.  ");
  paintchar('*');
  termrx('*');
  printf(" cbg %i.\n\n", version);
  while (1)
  {
    if (holding)
    {
      if (scrolllock == 0) holding = 0;
    }
    else
    {
      char c = sgetc(0) & 127;  // get char sans hang.
      if (c != 0) termrx(c);
      else if (locked==0) custime();
    }

    if (locked) kyread();
    else termtx();

    if (cmdreq) localcmd();
  }
}

sysreq(shift, control, alternate) char shift, control, alternate;
{ 
  if (shift) { clrhome(); return; }
  if (alternate) main();
  if (control)  cmdreq = 1;
}

localcmd()
 { cmdreq = 0;
   { char c;
     printf("\nEnter baud letter ->");
     c = kybdrd(1);
     if (c <= ' ') return; // ignore null response
     printf("%c\n", c);
     setbaud(c);
   }
}

termtx()
{ char c = kybdrd(0);  // Read keyboard with no hang.
  int i;
  if (c) switch(c)
   { default:
       sputc(c); // output on line interface
       for (i=0; i<1000; i++) continue;
       break;
     case 255: sputc(0); break;
   }
}

int readdec()
{ int r = 0;
  char c = sgetc(1) & 127;
  while(1)
  { if (c < '0') break;
    if (c > '9') break;
    r = (((r<<2)+r)<<1)  + c - '0';
    c = sgetc(1) & 127;
  }
  exitchar = c;
  return r;
}

int read8()
 { int r = read2();
   int s = read2();
   int t = read2();
   int u = read2();
   return (((((r << 8) + s) << 8) + t) << 8) + u;
 }

int read2()
 { int r = read1();
   return read1() + (r<<4);
 }

int read1()
 { int x = (sgetc(1) & 127) - '0';
   return (x > 9) ? x-7 : x;
 }

paintx2(i)
{ paintx1(i>>4);
  paintx1(i);
}

paintx1(i) char i;
{ i = i & 0x0F;
  termrx((i > 9) ? i+55: i+48);
}


sawrch(c) char c; /* needed by (e/skernel) */
{
  termrx(c);
}

termrx(c) char c;
{
  int i;
  cusout();
  if (escpending)
  { 
    switch(c)
    {
    case '[':
      { 
        int ny = readdec();
        int nx;
        switch (exitchar)
        {
        case 'A':  cursorup(); break;
        case 'B':  cursordown(); break;
        case 'C':  cursorright(); break;
        case 'D':  cursorleft(); break;
        case 'H':  justhome(); break;
        case 'J':  clrhome(); break;
        case 'K':  eraseeol(); break;
        case ';':
          nx = readdec();
          switch (exitchar)
          {
          case 'H':
            cusy =(((ny << 2) + ny) << 1)+topmarg;   // lineheight of 10
            cusx = ((nx-1) << 3) + leftmarg;
            notmoved = 0; /* We have moved */
            break;

          case 'M': move(nx, ny); break;
          case 'D': draw(nx, ny); break;
          }
          break;
        case 'm': mode = ny; break;
        }
      }
      break;
    case 'V': printf(" version %i ", version); break;
    case 'C': ( (addr))(); break;
    case 'B': addr = (char *) read8(); break;
    case 'P': *addr = read2(); addr++; break;
    case 'G': escpending = 0; paintx2(*addr); addr++; break;
    case 'T': escpending = 0; sawrch(*addr); addr++; break;
    case 'X': for (i=0; i<8; i++)  *(addr++) = sgetc(1); break;
    case 'E':
      while (1)
      {
        char ec = sgetc(1);
        sputc(ec);
        if (ec == 13) break;
      }
      break;
    }
    escpending = 0;
    return;
  }
  else if (graphpend)
  {
    c = ((c & 31) + 0x80);
    graphpend = 0;
 }

 if (cusx > wcols)
 {
   cusx = curmargin;
   cursordown();
 }

 if (c & 0xE0) paintchar(c);
 else
 switch(c)                                // Non escape codes.
 {
   case 7: bell(); break;
   case 8: cursorleft(); break;
   case 9: cursortab(); break;
   case 11: cursorup(); break;
   case 13: notmoved=0; cusx = curmargin; break;
   case 10: notmoved=1; cursordown(); break;
   case 12: clrhome(); break;
   case 27: escpending = 1; break;
   case 29: graphpend  = 1; break;
   case Xon: locked = 0; lockmark(0); break;
   case Xoff: locked = 1; lockmark(1); cusout();
 }
}

cursordown()
{
  if(cusy < cusymax) cusy = cusy + lineheight;
  else
  {
    cusout();
    scroll();
    if (scrolllock) holding = 1;
    cusy = cusy + lineheight - 32; // Scroll is of 32 points
  }
}



clrhome()
{ 
  crtcclr();
  move(homex, homey);
  lockmark(locked);
  curmargin = leftmarg;
  mode = 0;
  notmoved = 1;
}

justhome()
{ 
  move(homex, homey);
  lockmark(locked);
  curmargin = leftmarg;
  mode = 0;
  notmoved = 1;
}


