   /*------------------------------------------------------------------*
    *    + + +          CBG SOLOS OPERATION SYSTEM            + + +    *
    *                            (C) 1992                              *
    *                       David Greaves, cbg.                        *
    *------------------------------------------------------------------*/




/* cbg solos dshell : David's shell for solos */

#include "solosio.h"
#include "kmalloc.h"
#include "ksolos.h"
#include <fcntl.h>

const jobnstringlen = 32;

struct shelljob {
   int tid, jid;
   char namestring[jobnstringlen];
   char jobber;
   struct shelljob *next;
} *shelljobs;



int currentjobn = 0;
struct shelljob *newjob();

extern hostcmd();

#define TRC(X) 
#define SHELLTRC(X, Y) 
#define maxarg  (150)
#define maxline (4096)
#define pbuflen (maxline + 2)


char ** parse(char * buf);
char *pbuf;
char fname[maxline+2];
char **sargv;
char **parse1(p) char *p; /* Prototype */
extern int sputc();
char **parse2(p) char *p;

extern char signals;
struct shelljob *jcodefind(n);
extern int  sgetc();

extern int copycmd();

extern void readline();
extern wdos_dir();

extern int chdir();
extern int filedesexec();
int shellscript();
int shellsubproc();
extern int exists();


sayprompt()
{
  char *p = getenv("PROMPT");
  printf("  %s %s:", p, getenv("CDIR"));
}


/* Shellline is the input to the shell when a line is ready */
int shellline(buf) char * buf;
{
  int rc;
  char *p = buf;
  char **vv;
  if (buf == 0 || *buf == 0 || *buf == '#') return 0;

  while (*p && *p != '\n') p++;
  *p = (char) 0;
  vv = parse(buf);
  if (vv == (char **) 0)
  {
    printf("cannot parse %s\n", buf);
    return 1;
  }
  if (*vv == NULL) return 0;
  rc = shell1(vv);
  return rc;
}

/* Process any built in commands which run on the shell's own thread */
int shell1(vv) char **vv;
{
  if (strcmp(vv[0], "sipmon")==0) return sipmon(vv);
  if (strcmp(vv[0], "jobs")==0) return listjobs();
  if (strcmp(vv[0], "source")==0 || strcmp(vv[0], ".")==0) return source(vv[1]);
  if (strcmp(vv[0], "cd")==0)  return shellcd(vv[1]);
  if (strcmp(vv[0], "printenv")==0) return printenv(vv[1]);
  if (strcmp(vv[0], "sipcat")==0) return sipcat(&(vv[1]));
  if (strcmp(vv[0], "setenv")==0) 
  {
     setenv(vv[1], vv[2]);
     return 0;
  }
  if (strcmp(vv[0], "unsetenv")==0)
  {
    unsetenv(vv[1]);
    return 0;
  }
  if (*vv[0] == '%')
  {
    int jobn = atoi(vv[0]+1);
    struct shelljob *j = jcodefind(jobn);
    if (j == NULL) 
    {
      printf("No such job %i\n", jobn);
      return -1;
    }
    return foreground(jobn);
  }
  if (strcmp(vv[0], "fg")==0)
  {
    if (jcodefind(currentjobn) == NULL)  /* Try job 1 if in doubt */
                  currentjobn = 1;
    return foreground(currentjobn);
  }
  if (strcmp(vv[0], "bg")==0) return background(currentjobn);

  return shell2(vv);
}

/*
  Execute a line and check whether there's an ampersand on the end.
*/
int shell2(vv) char **vv;
{
  char **p = vv;
  struct thread * joni = newthread();
  int rc = 0;
  int nowaitp = 0;
  int jcode;
  struct shelljob *j;
  while (*p)
  { 
    if (**p == '&')
    {
      *p = 0;
      nowaitp = 1;
      break;
    }
    p++;
  }  

  jcode = freejobcode();
  j = newjob(jcode, joni, vv);

  startthread(joni, shellsubproc, vv, jcode);
  if (nowaitp)
  {
    background(jcode);
    /* nothing */
    return 0;
  }
  else 
  {
    return foreground(jcode);
  }
}


/*
 Execute a list of ; separated commands. run with a private thread.
*/
int shellsubproc(jcode)
{
  int rc = 0;
  char **vv = currentthread->vv;
  struct shelljob *j;
  SKIP(kprintf(std, "subproc %i %s\n", jcode, vv[0]));
  while(rc == 0)
  {
    char *t;
    char **start = vv;
    while(*vv && **vv != ';') vv++;
    t = *vv;
    *vv = 0;
    rc += shellsub1(start);
    *vv = t;
    if (t == NULL) break;
    vv ++;
  }
  j = jcodefind(jcode);
  if (j)
  {
    if (rc) fprintf(stderr, "[%i] %s exited rc = %i\n", jcode, j->namestring, rc);
    else if (j->jobber == 'B') fprintf(0, "[%i]  Done %s\n", jcode, j->namestring);
    freeshelljob(jcode);
  }
  return rc;
}


shellcd(v) char *v;
{
  int r = chdir(v);
  if (r==0) setenv("CDIR", v);
  return r;
}


/* Routine to execute a single command, possibly including input and
   output redirection or piping of two sub commands
*/
int shellsub1(vv) char **vv;
{
  int rc;
  char **p;
  int tracing = 0;
  p = vv;
  SHELLTRC("shell tok %s\n", *vv);
  signals = 0;
  while (*p)
  {
    int argstoskip = 0;
    if (strcmp(*p, "|")==0)
    {
      int fd1, fd2, fdextra;
      int rc;
      char **nt = p+1;
      struct thread *pth = newthread();
      *p = NULL;
      pipe(&fd2);
      currentthread->tstdout = fd1;
      pth->tstdin = fd2;
      rc = startthread(pth, shellsubproc, nt, 00);
      break;
    }
    else if (strcmp(*p, "trace")==0)
    {
      tracing = 1;
      argstoskip = 1;
    }
    else if (strcmp(*p, ">")==0)
    {
      int os = open(p[1], O_WRONLY, 0);
      if (os <= 0) return perror("failed to open output file");
      currentthread->tstdout = os
      argstoskip = 2;
    }
    else if (strcmp(*p, "<")==0)
    {
      int os = open(p[1], O_RDONLY, 0);
      if (os <= 0) return perror("cant failed to open input file");
      currentthread->tstdin = os;
      argstoskip = 2;
    }
    if (argstoskip)
    { 
      char **dp = p;
      char **sp = p + argstoskip;
      while (*sp) *dp++ = *sp++;
      *dp = NULL;      
    }
    else p++;
  }
  rc = shellev(vv, tracing);
  return rc;
 }

extern int execute();
extern int rename();
extern int cpcmd();

int shellev(vv, tracing) char** vv;
{
  int ss;
  char c;
  char * cmd;
  int tid;
  SHELLTRC("shelleval >%s< ", vv[0]);

  if (tracing) 
  {
    char **p = vv;
    while (*p) kprintf(std, "%s ", *p++);
    kprintf(std, "\n");
  }
  if (vv == NULL || *vv == (char) 0) return 0;
  if (strcmp(vv[0], "sipmv")==0) return rename(vv[1], vv[2]);
  if (strcmp(vv[0], "echo")==0)
  {
    char **s = &(vv[1]);
    while (*s) printf("%s ", *(s++));
    printf("\n");
    return 0;
  }

  if (strcmp(vv[0], "sipcp")==0) return copycmd(countargc(vv), vv);

  if (strcmp(vv[0], "sipdir")==0) return wdos_dir(vv[1]);


/*
  if (strcmp(vv[0], "fget")==0) return floppyget(vv[1]);
*/
  if (strcmp(vv[0], "logout")==0) return logout;


  SKIP(if (strcmp(vv[0], "get")==0) return getcmd(&(vv[1])));

  if (strcmp(vv[0], "if")==0)
  {
    if (strcmp(vv[1], "-e")==0)
        { if (exists(vv[2])) return shellev(&vv[3], tracing);
          return 0;
        }
    if (strcmp(vv[1], "!-e")==0)
        { if (exists(vv[2])==0) return shellev(&vv[3], tracing);
          return 0;
        }
     return 0;
  }
  cmd = vv[0];
  while (*cmd == ' ') cmd++;
  if (*cmd == (char) 0) return 0;
  ss = openpath(cmd, getenv("PATH"));
  if (ss == 0)
  {
    printf("cannot load %s ", vv[0]);
    perror("");
    return 1;
  }
  c = read1(ss);

  if (c != '#' && c != ':')
  {
    close(ss);
    kprintf(std, "bad magic %x, cannot execute\n", c);
    return 1;
  }

  if (c == ':')  return filedesvv(vv, ss);
  if (c == '#')  shellscript(vv, ss, tracing);
}

#define scriptsize (4096)


int shellscript(vv, ss, tracing)  char **vv;
{
  char * lpt, *buf;
  char * scriptend;
  char * lhs = "1";  /* First arg is called "$1" */;
  char **v = vv + 1;
  strcpy(lhs, "1");
  while(*v)  /* Add args to environment */
  {
    setenv(lhs, *v++);
    (*lhs) ++;  /* Not re-entrant self modding code here ! */
  }
  return sccom(ss, tracing)
}

int source(f) char *f;
{
  int fd = open(f, O_RDONLY, 0);
  if (fd>0) return sccom(fd, 0);
  else return perror("source: ");
}


int sccom(ss, trace)
{
  char *scriptbuffer = kmalloc(scriptsize, "SCRIPT");
  char *lpt;
  char *scriptend;
  int chars;

  chars = read(ss, scriptbuffer, scriptsize);
  close(ss);
  scriptend = scriptbuffer+chars;
  lpt = scriptbuffer;
  while (lpt < scriptend)
  {
    if (lpt[0] == '\\' && lpt[1] == '\n')
    {
      lpt[0] = ' ';
      lpt[1] = ' ';
    }
    lpt ++;
  }
  lpt = scriptbuffer;
  while (lpt < scriptend)
  {
    int rc;
    char *buf = lpt;
    if (trace)
    {
      char *j = lpt;
      while (*j != '\n') kputchar(*j++);
      kputchar('\n');
    }
    while(*(lpt++) != '\n') continue;
    rc = shellline(buf);
    if (rc) return rc;
  }
  return 0;
}


/* Create a copy of the environment */
struct envitem *copyenv(p) struct envitem *p;
{
  struct envitem *r;
  if (p == NULL) return NULL;
  r = (struct envitem *) kmalloc(sizeof(struct envitem), "COPYE");  
//  kprintf(std, "Copying %x %s\n", p, p->etag);
  r->next = copyenv(p->next);
  r->etag = kmalloc (strlen(p->etag)+1, "CPETAG");
  strcpy(r->etag, p->etag);
  r->eval = kmalloc (strlen(p->eval)+1, "CPEVAL");
  strcpy(r->eval, p->eval);
  return r;
}

freeenv(p) struct envitem *p;
{
  while (p)
  {
    struct envitem *pn = p->next;
    kfree(p->etag);
    kfree(p->eval);
    kfree(p);
    p = pn;
  }
}

void setenv(char *tag, char *value)
{
  struct envitem *o = currentthread->env;
  struct envitem *new;
  char *ntag, *nval;
  if (tag == NULL) return;
  if (value == NULL) value = "";
  while(o)
  {
    if (strcmp(o->etag, tag)==0) break;
    o = o->next;
  }
  if (o)
  {
    new = o;
    kfree(o->eval);
  }
  else 
  {
    new = (struct envitem *) kmalloc(sizeof(struct envitem), "SETENV");
    ntag = kmalloc(strlen(tag)+1, "SETENV");
    new->etag = ntag;   
    strcpy(ntag, tag);
    new->next = currentthread->env;
    currentthread->env = new; 
  }

  nval = kmalloc(strlen(value)+1, "SETENV");
  strcpy(nval, value);
  new->eval = nval;
}

int printenv(char *v) 
{
  struct envitem *oo =  currentthread->env;
  char *p = getenv(v);
  if (p) kprintf(std, "%s=%s\n", v, p);
  else while(oo)
  {
    kprintf(std, "%s=%s\n", oo->etag, oo->eval);
    oo = oo->next;
  }
  return 0;
}

char * getenv(tag) char * tag;
{
  struct envitem *o = currentthread->env;
  if (tag == NULL) return NULL;
  while(o)
  {
    SKIP(kprintf(std, "getenv %x\n", o));
    if (strcmp(o->etag, tag)==0) return o->eval;
    o = o->next;
  }
  return NULL;
}

unsetenv(tag)
{
  struct envitem **optr = &(currentthread->env);
  while(*optr)
  { 
    if (strcmp((*optr)->etag, tag) == 0)
    {
      struct envitem *v = *optr;
      *optr = (*optr)->next;
      kfree(v->etag);
      kfree(v->eval);
      kfree(v);
      return;
    }
    optr = &((*optr)->next);
  }
}

char parsetag[maxline];

/* Expand a star in the current directory */
char *sjvec = 0;
char * starjob(mixer)
{
  if (sjvec == NULL) sjvec = kmalloc(4096, "STAR");
  starslist(sjvec, mixer);
  return sjvec;
}



char **parse(p) char *p;
{
  char *d = pbuf;
  int ddone = pbuflen;
  int i;
  SHELLTRC("parse >%s< ", p);
  while(*p == ' ') p++;
  while (*p)
  {
    if (*p == '$')              /* $ substitution */
    {
      char *tt = parsetag;
      p++;
      while (isupper(*p) || isdigit(*p))
            { *(tt++) = *(p++);
            }
      *tt = (char) 0;
      tt = getenv(parsetag);
      if (tt) while (*tt)
      { *(d++) = *(tt++);  /* do not copy trailing null */
        if ((ddone--) == 0) return NULL;
      }
      continue;
    }
    if (*p == ';')
    {
      *d++ = ' ';
      *d++ = *p++;
      *d++ = ' ';
      ddone -= 3;
    }
    else
    {
      *d++ = *p++;
      ddone -= 1;
    }
    if (ddone <= 0) return NULL;
  }
  *d = (char) 0;
  return parse1(pbuf);
}


char **  parse1(p) char *p;              /* split at space boundaries */
{                                   /* fill argv and set argc */
   char **narg = sargv;
   int i;
   int argc = 0;
   SHELLTRC("parse1 >%s< ", p);
   while(*p == ' ') p++;
/*
   for (i=0; i<maxarg; i++) argv[i] = "";
*/
   while (*p)
    {
      argc++;
      if (argc >= maxarg) break;
      if (*p == '\'')
      {
        p++;
        *narg = p;
        while (*p != '\'' && *p != (char) 0) p++;
      }
      else
      {
        if (*p == '>' || *p == '<')  /* Add spaces around redirection operators */
        {
          *narg = (*p == '>') ?  ">" : "<";
          p++;
        }
        else
        {
          *narg = p;
          while(*p != ' ' && *p != (char) 0) p++;
        }
      }
      narg++;
      *narg = 0;  /* Terminate arg list as required by execve */
      if (*p != 0) *(p++) = 0;
      while(*p == ' ') p++;
    }
   return parse2(sargv);
 }

char **parse2(vv) char **vv;  /* Parse 2 expands * as filenames */
{
  char **p = vv;
  p--;
  while(1)
  {
    char *q, *r;
    int files;
    p++;
    if (*p == NULL) return vv;  /* Finished */
    if (**p == '\'' || **p == '\"') continue;  /* Skip entries with quotes */
    q = *p;
    while (*q && *q != '*') q++;
    if (*q == (char) 0) continue;
    r = starjob(*p);
    files = 0;
    q = r;
    while(*q)  /* Split list of file names into null separated list and count */
    {
      files ++;
      while (*q && *q != ' ') q++;
      if (*q == 0) break;
      *q++ = 0;
    }
    SKIP(kprintf(std, "files %i of %i\n", files, maxarg));
    if (files + countargc(vv) >= maxarg) 
    {
      printf("shell: too many items on line\n");
      return NULL;
    }

    /* Move rest of argv up by files-1 positions. */

    if (files > 0)
    {
      char **dd, **ss = p;
      while (*ss) ss++;
      dd = ss + (files - 1);
      dd[1] = NULL;
      while (ss != p) *dd-- = *ss--;

      /* Copy in the file names from starjob */
      dd = p;
      while (files > 0)
      {
        *dd++ = r;
        r = r + strlen(r) + 1;
        files --;
      }
    }
    else *p = ""; /* No files matched case */
  }
}

int  exec(string) char *string;
{
  int rc;
  rc = shellline(string);
  if (rc) printf("exec: Exit  %i   %s\n", rc, string);
  return rc;
}

/* Search a path for a string */
int openpath(name, path) char *name, *path;
{
  int rc;
  char *pr =
  "----------------------------------------------------------------------";
  if (*name == '/' || *name == '.') return open(name, O_RDONLY);  /* explicit path name used */
  while (*path)
  {
    int c = 0;
    char * d = pr;
    char * n = name;
    while (c<60 && *path && *path != ':')
    {
      *(d++) = *(path++);
      c++;
    }
    if (*path) path++; /* skip colon */
    if (c) *(d++) = '/';
    while (c<60 && *n)
    {
      *(d++) = *(n++);
      c++;
    }
    *d = (char) 0;
    rc = open(pr,  O_RDONLY);
    if (rc>0) return rc;
  }
  return 0;
}

listjobs()
{
  struct shelljob * t = shelljobs;
  while(t)
  {
    if (t->jobber != 'Z')
    kprintf(std, "[%i]  %c %32s    tid=%i\n", 
         t->jid,
         (t->jid == currentjobn) ? '*' : ' ',
         t->namestring, t->tid);

    t = t-> next;
  }
  return 0;
}

freeshelljob(jcode)
{
  struct shelljob *j;
  if (j = jcodefind(jcode)) j->jobber = 'Z';
}


int freejobcode()
{
  int i = 1;
  while (jcodefind(i)) i++;
  return i;
}



/* Find a job by job number */
struct shelljob *jcodefind(n)
{
  struct shelljob *r = shelljobs;
  while (r) 
  {
    if (r->jid == n && r->jobber != 'Z') break;
    r = r->next;
  }
  return r;
} 

/* Add a new job to the job list */
struct shelljob * newjob(jcode, joni, vv) 
struct thread *joni;
char **vv;
{
  struct shelljob *j = shelljobs;
  while (j)
  {
    if (j->jobber == 'Z') break;
    j = j->next;
  }
  if (j == NULL)
  {
    j = (struct shelljob *) malloc(sizeof(struct shelljob));
    j->next = shelljobs;
    shelljobs = j;
  }
  j->jobber = 'V';
  j->jid = jcode;
  strncpy(j->namestring, vv[0], jobnstringlen);
  j->tid = joni->tid;
  return j;
}

shell_init()
{
  sargv = (char **) kmalloc(maxarg * sizeof(char **), "ARGV");
  pbuf = kmalloc(pbuflen, "PBUF");
  shelljobs = NULL;
  currentjobn = 0;
}

int foreground(jcode)
{
  struct shelljob *j;
  SKIP(printf("Foreground for jcode %i\n", jcode));
  j = jcodefind(jcode);
  if (j)
  {
    int rc;
    struct thread *t = tid2thread(j->tid);
    if (t) t->status = 'R';
    currentjobn = jcode;
    while (1)
    {
      rc = jt_wait();
      SKIP(printf("Shell awoken %i\n", rc));
      t = tid2thread(j->tid);  /* See if thread no longer exists */
      if (t == NULL) break;
      SKIP(printf("shell: signalled with status %c\n", t->status));
      if (t->status == 'S') break;  /* if stopped by ^Z */
    }
    return rc;
  }
  else return -1;
}

background(jcode)
{
  struct shelljob *j;
  if (j = jcodefind(jcode))
  {
    struct thread *t = tid2thread(j->tid);
    j->jobber = 'B';  /* Mark as background so that rc is printed on exit */
    if (t) t->status = 'R';  /* Running job */
    return 0;
  }
  else return -1;
}

/* end of dshell.c */


